package com.alterdekim.flash.decompiler.mapper;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Map;

@Slf4j
public class FlashMapper {

    public static <T> T mapToObj(Map<String, Object> m, Class<T> obj) {
        try {
            T result = obj.getDeclaredConstructor().newInstance();
            Arrays.stream(obj.getDeclaredFields())
                    .filter(f -> m.containsKey(f.getName()) || (f.isAnnotationPresent(FlashField.class) && m.containsKey(f.getAnnotation(FlashField.class).name())))
                    .map(f -> Pair.of(f, f.isAnnotationPresent(FlashField.class) ? f.getAnnotation(FlashField.class).name() : f.getName()))
                    .forEach(p -> {
                        try {
                            Field f = p.getLeft();
                            String name = p.getRight();
                            f.setAccessible(true);
                            if( f.isAnnotationPresent(FlashField.class) && f.getAnnotation(FlashField.class).useCustomConverter()) {
                                f.set(result, f.getAnnotation(FlashField.class).converter().getDeclaredConstructor().newInstance().convert(m.get(name)));
                                return;
                            }
                            f.set(result, m.get(name));
                        } catch (IllegalAccessException | NoSuchMethodException | InstantiationException |
                                 InvocationTargetException e) {
                            log.error("FlashMapper valueSet error: {}", e.getMessage());
                        }
                    });
            return result;
        } catch (Exception e) {
            log.error("FlashMapper error: {}", e.getMessage());
        }
        return null;
    }
}
