package com.alterdekim.flash.decompiler.translator;

import com.alterdekim.flash.decompiler.action.ActionPush;
import com.alterdekim.flash.decompiler.item.*;
import com.alterdekim.flash.decompiler.action.ByteCodeAction;
import com.alterdekim.flash.decompiler.item.internal.GetMemberItem;
import com.alterdekim.flash.decompiler.item.internal.GetVariableItem;
import com.alterdekim.flash.decompiler.item.internal.InitArrayItem;
import com.alterdekim.flash.decompiler.item.internal.InitObjectItem;
import com.alterdekim.flash.decompiler.tag.DoAction;
import lombok.extern.slf4j.Slf4j;

import java.util.*;

@Slf4j
public class Flash2Java {

    private final List<String> pool;
    private final List<ByteCodeAction> actions;
    private final Stack<ByteCodeItem> stack;
    private final Map<String, Object> vars;

    public Flash2Java(DoAction doAction) {
        this.pool = doAction.getActionPool();
        this.stack = new Stack<>();
        this.actions = doAction.getActions();
        this.vars = new HashMap<>();
    }

    public Map<String, Object> convert() {
        for (ByteCodeAction action : actions) {
            switch (action.getType()) {
                case DefineLocal2:
                    // doesn't need implementation due to HashMap's properties. (variableName)
                    this.stack.pop();
                    break;
                case DefineLocal:
                    defineLocal();
                    break;
                case GetVariable:
                    getVariable();
                    break;
                case SetMember:
                    setMember();
                    break;
                case Add:
                    reduceViaAdder();
                    break;
                case InitObject:
                    initObject();
                    break;
                case InitArray:
                    initArray();
                    break;
                case Push:
                    this.stack.addAll(((ActionPush) action).getItems());
                    break;
                case SetVariable:
                    setVariable();
                    break;
                case GetMember:
                    getMember();
                    break;
            }
        }
        return this.vars;
    }

    private void initObject() {
        int numberOfEntries = (int) Double.parseDouble(this.stack.pop().getValue()+""); // harsh way todo: rewrite
        HashMap<String, Object> obj = new HashMap<>();
        for( int i = 0; i < numberOfEntries; i++ ) {
            Object val = valConverter( this.stack.pop() );
            String key = valToString( this.stack.pop() );
            obj.put(key, val);
        }
        this.stack.push(new InitObjectItem(obj));
    }

    private void initArray() {
        int numberOfElements = (int) Double.parseDouble(this.stack.pop().getValue()+""); // harsh way todo: rewrite
        HashMap<String, Object> obj = new HashMap<>();
        for( int i = 0; i < numberOfElements; i++ ) {
            Object val = valConverter( this.stack.pop() );
            obj.put(String.valueOf(i), val);
        }
        this.stack.push(new InitArrayItem(obj));
    }

    private void reduceViaAdder() {
        String s1 = valToString( this.stack.pop() );
        String s2 = valToString( this.stack.pop() );
        this.stack.push(new StringItem(s2 + s1));
    }

    private void getMember() {
        String property = valToString( this.stack.pop() );
        ByteCodeItem val = this.stack.pop();
        if( val.getType() == ActionPushType.GET_MEMBER ) {
            ArrayDeque<String> l = (ArrayDeque<String>) val.getValue();
            l.add(property);
            this.stack.push(new GetMemberItem(l));
            return;
        }
        String key = (String) val.getValue();
        this.stack.push(new GetMemberItem(new ArrayDeque<>(List.of(key, property))));
    }

    private void setMember() {
        ByteCodeItem val = this.stack.pop();
        String index = valToString( this.stack.pop() );
        ByteCodeItem varOrMember = this.stack.pop();
        if( varOrMember.getType() == ActionPushType.GET_VARIABLE ) {
            Map<String, Object> m = (Map<String, Object>) valConverter(varOrMember);
            m.put(index, valConverter(val));
            return;
        }
        Map<String, Object> m = recursiveGet((ArrayDeque<String>) varOrMember.getValue());
        if( val.getType() == ActionPushType.GET_VARIABLE ) {
            m.put(index, this.vars.get(valToString(val)));
            return;
        }
        m.put(index, valConverter(val));
    }

    private Map<String, Object> recursiveGet(Deque<String> l) {
        Map<String, Object> m = this.vars;
        for (String s : l) {
            m = (HashMap<String, Object>) m.get(s);
        }
        return m;
    }

    private String valToString(ByteCodeItem item) {
        switch (item.getType()) {
            case POOL_INDEX:
                return fromPool( (Integer) item.getValue() );
            case STRING:
                return (String) item.getValue();
            case GET_VARIABLE:
                return (String) vars.get((String) item.getValue());
            default:
                return String.valueOf( item.getValue() );
        }
    }

    private Object valConverter(ByteCodeItem item) {
        return switch (item.getType()) {
            case POOL_INDEX -> fromPool((Integer) item.getValue());
            case GET_VARIABLE -> vars.get((String) item.getValue());
            default -> item.getValue();
        };
    }

    private void setVariable() {
        Object val = valConverter( this.stack.pop() );
        String variableName = valToString( this.stack.pop() );
        this.vars.put(variableName, val);
    }

    private void getVariable() {
        String variableName = valToString( this.stack.pop() );
        this.stack.push(new GetVariableItem(variableName));
    }

    private void defineLocal() {
        ByteCodeItem val = this.stack.pop();
        String name = valToString(this.stack.pop());
        this.vars.put(name, valConverter(val));
    }

    private String fromPool(int index) {
        return pool.get(index);
    }
}